package org.ginsim;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.List;

import org.colomoto.logicalmodel.services.Colomoto;
import org.colomoto.logicalmodel.services.ExtensionLoader;
import org.ginsim.common.application.CurrentOS;
import org.ginsim.common.application.LogManager;
import org.ginsim.common.application.OSXAdapter;
import org.ginsim.common.application.OptionStore;
import org.ginsim.common.application.Txt;
import org.ginsim.core.service.ServiceClassInfo;
import org.ginsim.commongui.dialog.GUIMessageUtils;
import org.ginsim.commongui.utils.ImageLoader;
import org.ginsim.core.graph.GraphFactory;
import org.ginsim.core.graph.GraphManager;
import org.ginsim.core.graph.objectassociation.ObjectAssociationManager;
import org.ginsim.core.service.ServiceManager;
import org.ginsim.gui.GUIManager;
import org.ginsim.gui.service.ServiceGUIManager;
import org.ginsim.gui.shell.AboutDialog;


/**
 * Simple launcher: parse command line arguments, show some help, open files or run the script mode accordingly.
 * 
 * @author Aurelien Naldi
 */
public class Launcher {

	/**
	 * True if the developers extras (toolkit menu, under-developement...) should be enabled
	 */
	public static boolean developer_mode = false;
	
	/**
	 * The main "main" function, launching GINsin GUI or script handler.
	 * 
	 * @param args
	 * @throws InstantiationException 
	 * @throws IllegalAccessException 
	 * @throws ClassNotFoundException 
	 */
	public static void main(String[] args) throws ClassNotFoundException, IllegalAccessException, InstantiationException {
        ExtensionLoader.loadExtensions("extensions", Launcher.class);
        init();
		
        List<String> open = new ArrayList<String>();

        /*
         * parse arguments:
         *  - open files
         *  - script mode
         *  - help
         */
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-lm")) {
                String[] lmargs = new String[args.length-(i+1)];
                System.arraycopy(args, i+1, lmargs, 0, lmargs.length);
                Colomoto.main(lmargs);
                return;
            } else if (args[i].equals("-s")) {
        		ScriptLauncher script = new ScriptLauncher();
                if (args.length == i+1) {
                	script.help();
                    return;
                }
                
                i++;
                String[] pyargs = new String[args.length-(i+1)];
                System.arraycopy(args, i+1, pyargs, 0, pyargs.length);
                script.run(args[i], pyargs);
                
                // force exit as jython seems to often delay it
                System.exit(0);
                return;
            } else if (args[i].equals("-n")) {
        		open.add(null);
        		continue;
            } else if (args[i].equals("--dev")) {
                developer_mode = true;
                continue;
            } else if (args[i].equals("--devinfo")) {
                developer_mode = true;
                devInfo();
                return;
            } else if (args[i].startsWith("-")) {
            	if (!args[i].equals("-h")) {
                    System.out.println("Unknown option: "+args[i]);
            	}
                System.out.println("Available options:");
                System.out.println("\t<file>: open <file> on startup.");
                System.out.println("\t-n: start with a new regulatory graph.");
                System.out.println("\t-s: display the script help message.");
                System.out.println("\t-s <file>: run \"script\" from <file>. Extra arguments are script arguments.");
                System.out.println("\t-h:  display this message.");
                System.out.println("\t--dev : enable the developer's options.");
                System.out.println("\t-lm: LogicalModel mode: takes the same arguments as the LogicalModel conversion tool.");
                return;
            }
        	
            File f = new File(args[i]);
            if (f.exists()) {
                open.add(args[i]);
            } else {
            	LogManager.error("Required file does not exist: " + f);
            }
        }
		
		initGUI();
		GUIManager.getInstance().startup( open);
	}

	/**
	 * Initialise GINsim core: detect the current directory, load services
	 */
	private static void init() {
		String basedir = null;
		
		Class<?> cl = Launcher.class;
		String clname = cl.getName().replace(".",	"/") + ".class";
		String path = cl.getClassLoader().getResource(clname).toString();
		if (path.startsWith("file:")) {
			basedir = path.substring(5,  path.length() - clname.length());
		} else if (path.startsWith("jar:file:")) {
			File jar = new File(path.substring(9,  path.length() - clname.length() - 2));
			basedir = jar.getParent();
        }

        if (basedir != null) {
            try {
                basedir = URLDecoder.decode(basedir, "UTF-8");
            } catch (Exception e) {
                basedir = null;
            }
        }

        if (basedir == null) {
            basedir = System.getProperty("user.dir");
        }

		try{
			LogManager.init( basedir, 2, false);
		}
		catch( IOException io){
			System.out.println("TestRefactor.main() : Unable to initialize the debugger");
		}

        // extend the classpath if needed
        File extensionDir = new File(basedir, "extensions");
        if (extensionDir.isDirectory()) {
            FileFilter filter = new FileFilter() {
                public boolean accept(File file) {return file.getPath().toLowerCase().endsWith(".jar");}
            };
            File[] files = extensionDir.listFiles();
            try {
                URL[] urls = new URL[files.length];
                int i=0;
                for (File f: files) {
                    urls[i++] = f.toURI().toURL();
                }

                URLClassLoader ucl = new URLClassLoader(urls);
                // TODO: pass this class loader to the service managers
            } catch (IOException e) {
                LogManager.error("Could not load extension files");
            }
        }

		ServiceManager.getManager();
	}
	
	/**
	 * Initialisation method for GINsim GUI.
	 * This method will only load all required resources, it will not create the first window.
	 */
	private static void initGUI() {
		Txt.push("org.ginsim.messages");
		ImageLoader.pushSearchPath("/org/ginsim/icon");
		ImageLoader.pushSearchPath("/org/ginsim/icon/action");
		AboutDialog.setDOAPFile("/GINsim-about.rdf");
		try {
			OptionStore.init(Launcher.class.getPackage().getName());
		} catch (Exception e) {
			GUIMessageUtils.openErrorDialog(e, null);
		}
		
    	// register OSX callback if appropriate
        if (CurrentOS.CURRENT_OS == CurrentOS.MACOSX) {
            try {
                // Generate and register the OSXAdapter, passing it a hash of all the methods we wish to
                // use as delegates for various com.apple.eawt.ApplicationListener methods
            	GUIManager guiManager = GUIManager.getInstance();
                OSXAdapter.setQuitHandler(guiManager, guiManager.getClass().getDeclaredMethod("quit", (Class[])null));
                OSXAdapter.setAboutHandler(guiManager, guiManager.getClass().getDeclaredMethod("about", (Class[])null));
                OSXAdapter.setFileHandler(guiManager, guiManager.getClass().getDeclaredMethod("loadGINMLfile", new Class[] { String.class }));
            } catch (Exception e) {
                System.err.println("Error while loading the OSXAdapter:");
                e.printStackTrace();
            }
        }
    }

    /**
     * Print information for developers: list of loaded services
     */
    private static void devInfo() {

        // load everything
        GraphManager graphManager = GraphManager.getInstance();
        ObjectAssociationManager assocManager = ObjectAssociationManager.getInstance();
        ServiceManager srvManager = ServiceManager.getManager();
        ServiceGUIManager srvGUIManager = ServiceGUIManager.getManager();

        System.out.println("Graphs");
        for (ServiceClassInfo info: graphManager.getGraphsInfo()) {
            System.out.println(info);
        }
        System.out.println();

        System.out.println("Data handlers");
        for (GraphFactory factory: graphManager.getGraphFactories()) {
            Class cl = factory.getGraphClass();
            System.out.println("# "+cl.getSimpleName());
            for (ServiceClassInfo info: assocManager.getDataManagerInfo(cl)) {
                System.out.println(info);
            }
            System.out.println();
        }

        for (ServiceClassInfo info: assocManager.getDataManagerInfo(null)) {
            System.out.println(info);
        }
        System.out.println();

        System.out.println("Services");
        for (ServiceClassInfo info: srvManager.getServicesInfo()) {
            System.out.println(info);
        }
        System.out.println();

        System.out.println("GUI for services");
        for (ServiceClassInfo info: srvGUIManager.getServicesInfo()) {
            System.out.println(info);
        }
    }
 }
